import Pengyne
import random

engine = Pengyne.GetEngine()

class VoxelBoxState(Pengyne.GameState):
	def __init__(self):
		Pengyne.GameState.__init__(self)

		# change the camera to 3d
		engine.GetCamera().SetupView3D(90, 1, 100)
		
		# Block is below
		self.block = Block()
		self.block.GenerateBuffer()
		
		# Vertex buffers batch large amounts of geometry that render at the same time
		self.buffer = Pengyne.VertexBuffer()
	
		# Add the vertex definition, Normal is used for lighting
		self.buffer.AddMember("Position", Pengyne.GL.FLOAT, 3)
		self.buffer.AddMember("Normal", Pengyne.GL.FLOAT, 3)
		self.buffer.AddMember("Color", Pengyne.GL.UNSIGNED_BYTE, 4)
		
		# Add the block geometry to the buffer, see Block class below
		self.block.AddToBuffer(self.buffer)
		
		# create a simple shader. The lighting is currently wrong, but it works
		# for demostrational purposes
		self.shader = Pengyne.Shader()
		self.shader.VertexFromFile("VoxelBox.vert")
		self.shader.FragmentFromFile("VoxelBox.frag")
		# Don't forget to link
		self.shader.Link();
		
		# set the lighting values
		self.shader.Set("lightpos", [100,100,100])
		self.shader.Set("lightdiffuse", [1.0,1.0,1.0])

		# move the camera away from the block so we can see it
		cam = engine.GetCamera()
		cam.Move((0,0, -self.block.size[2]*2))
		
		# build the render op to draw
		self.op = Pengyne.BasicRenderOp()
		self.op.SetVertices(self.buffer)
		self.op.SetShader(self.shader)
	
		# load the menu
		self.menu = engine.GetGUIContext().LoadDocument("VoxelBox.rml")
		# Don't forget to show it
		self.menu.Show(True)
	def Render(self):
		# make the box rotate 
		self.op.Rotate((1,0,0), 15)
		self.op.Rotate((0,1,0), 25)
		self.op.Rotate((0,0,1), 35)
		
		#draw it
		engine.Render(self.op)
	
	def SetMode(self, name):
		# switch mode, from VoxelBox.rml
		# Triangles is what you want most of the time, 
		#  but Lines and Points are useful as well
		self.buffer.SetMode(name)
		
	def OnClose(self):
		engine.Quit()
		
	def OnClick(self, event):
		pass

	def OnKeyDown(self, action):
		if action == "Quit":
			engine.Quit()
			
	def OnKeyUp(self, action):
		pass
		
#helper that adds 2 float arrays
def add(l, r):
	return [a + b for a,b in zip(l, r)]

# the number of voxels of a block	
DEFAULTBLOCKSIZE = 16

# the sides of the cube. Cached here so it doesn't have to be created more than once
# (side of cube (normal), [square that makes up that side])
edgegeometry = (
	(-1,0,0, [(-.5, -.5, -.5), (-.5,  .5, -.5), (-.5,  .5,  .5), (-.5, -.5,  .5)]),
	( 1,0,0, [( .5, -.5, -.5), ( .5,  .5, -.5), ( .5,  .5,  .5), ( .5, -.5,  .5)]),
	(0,-1,0, [(-.5, -.5, -.5), ( .5, -.5, -.5), ( .5, -.5,  .5), (-.5, -.5,  .5)]),
	(0, 1,0, [(-.5,  .5, -.5), ( .5,  .5, -.5), ( .5,  .5,  .5), (-.5,  .5,  .5)]),
	(0,0,-1, [(-.5, -.5, -.5), ( .5, -.5, -.5), ( .5,  .5, -.5), (-.5,  .5, -.5)]),
	(0,0, 1, [(-.5, -.5,  .5), ( .5, -.5,  .5), ( .5,  .5,  .5), (-.5,  .5,  .5)]))


class Block:
	def __init__(self):
		r = random.Random()
		self.size = [DEFAULTBLOCKSIZE, DEFAULTBLOCKSIZE, DEFAULTBLOCKSIZE]
		self.palette = [
			# clear multiple times to make the resulting box more interesting
			(0,0,0,0),
			(0,0,0,0),
			(0,0,0,0),
			(0,0,0,0),
			(0,0,0,0),
			(255,0,0,255),
			(0,255,0,255),
			(0,0,255,255),
			(255,255,0,255),
			(0,255,255,255),
			(255,0,255,255),
			(255,255,255,255),
		]
		
		palettecount = len(self.palette)-1
		# init the data
		self.data = [[[r.randint(0,palettecount) for x in range(self.size[0])] for y in range(self.size[1])] for z in range(self.size[2])]
		
		# buffer is not generated yet
		self.buffer = []

	# walk the data values and add faces when a cube's neighbor is not fully opaque (alpha 255)
	def GenerateBuffer(self):
		self.buffer = []

		# the center of the cube, so that it is centered on screen
		half = [self.size[0] / 2, self.size[1] / 2, self.size[2] / 2]
		
		for x in range(self.size[0]):
			for y in range(self.size[1]):
				for z in range(self.size[2]):
					coloridx = self.GetColorIdxAt(x,y,z)
					# color must be a list, not a tuple so it may be appended later
					color = list(self.palette[coloridx])
					if color[3] != 0:
						for ox, oy, oz, vertices in edgegeometry:
							if self.IsPixelTransparent(ox+x, oy+y, oz+z):
								normal = [ox, oy, oz]
								# recenter the vertices of the face
								vertices = [add(v, (x - half[0],y - half[1],z - half[2])) for v in vertices]
								for idx in [0,1,3,1,2,3]:
									self.buffer += vertices[idx] + normal + color
		
	def IsPixelTransparent(self, x, y, z):
		if not self.IsPointInBlock(x,y,z):
			return True
		idx = self.GetColorIdxAt(x,y,z)
		return self.palette[idx][3] < 255
		
	def IsPointInBlock(self, x, y, z):
		if x < 0 or x >= self.size[0] or y < 0 or y >= self.size[1] or z < 0 or z >= self.size[2]:
			return False
		return True
		
	def GetColorIdxAt(self, x, y, z):
		return self.data[x][y][z]

	# set the data into the VertexBuffer passed in from the VoxelBoxState above
	def AddToBuffer(self, buffer):
		buffer.Add(self.buffer)
		

		